#include <strstrea.h>
#include <graphics.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <new.h>

const char *error1 = "invalid file format";
const char *error2 = "graphics not initialized";
const char *error3 = "unable to open file";
const char *error4 = "not enough memory";

int n = 0;

void Draw3DBox(int x, int y, int w, int h)
{
  setcolor(7);
  line(x, y, x + w - 1, y);
  line(x, y, x, y + h - 1);
  setcolor(15);
  line(x + 1, y + 1, x + 1, y + h - 3);
  line(x + 1, y + 1, x + w - 3, y + 1);
  setcolor(0);
  line(x, y + h - 1, x + w - 1, y + h - 1);
  line(x + w - 1, y, x + w - 1, y + h - 1);
  setcolor(8);
  line(x + 1, y + h - 2, x + w - 2, y + h - 2);
  line(x + w - 2, y + 1, x + w - 2, y + h - 2);
  setfillstyle(1, 7);
  bar(x + 2, y + 2, x + w - 3, y + h - 3);
}

void Draw3DHole(int x, int y, int w, int h, int color)
{
  setcolor(8);
  line(x, y, x + w - 2, y);
  line(x, y, x, y + h - 2);
  setcolor(0);
  line(x + 1, y + 1, x + w - 3, y + 1);
  line(x + 1, y + 1, x + 1, y + h - 3);
  setcolor(15);
  line(x + w - 1, y + h - 1, x, y + h - 1);
  line(x + w - 1, y + h - 1, x + w - 1, y);
  setcolor(7);
  line(x + w - 2, y + h - 2, x + 1, y + h - 2);
  line(x + w - 2, y + h - 2, x + w - 2, y + 1);
  setfillstyle(1, color);
  bar(x + 2, y + 2, x + w - 3, y + h - 3);
}

void DrawCube(int main_x, int main_y, int x, int y, int z, char color)
{
  int xx, yy;

  xx = main_x + x * 12;
  xx += 12 * y;
  yy = main_y - x * 4;
  yy += 4 * y;
  yy += z * 12;

  int poly1[] = {xx, yy, xx + 12, yy - 4, xx + 12,
                 yy + 8, xx, yy + 12, xx, yy};
  int poly2[] = {xx, yy, xx - 12, yy - 4, xx - 12,
                 yy + 8, xx, yy + 12, xx, yy};
  int poly3[] = {xx, yy, xx - 12, yy - 4, xx, yy - 8,
                 xx + 12, yy - 4, xx, yy};

  setcolor(0);
  if (y == 2)
  {
    setfillstyle(1, color);
    fillpoly(5, poly1);
  }
  if (!x)
  {
    setfillstyle(1, color);
    fillpoly(5, poly2);
  }
  if (!z)
  {
    setfillstyle(1, color);
    fillpoly(5, poly3);
  }
}

void DrawBigCube(int xx, int yy, char *s)
{
  int z;

  for (int y = 0; y < 2; y++)
    for (int x = 1; x < 3; x++)
    {
      int c = s[(2 - y) * 9 + x] - 'a' + 1;
      DrawCube(xx, yy, x, y, 0, c);
    }

  for (z = 2; z >= 0; z--)
    for (int y = 0; y < 2; y++)
    {
      int c = s[(2 - y) * 9 + z * 3] - 'a' + 1;
      DrawCube(xx, yy, 0, y, z, c);
    }

  for (z = 2; z >= 0; z--)
    for (int x = 2; x >= 0; x--)
    {
      int c = s[z * 3 + x] - 'a' + 1;
      DrawCube(xx, yy, x, 2, z, c);
    }
}

class TList {
    char s[28];
    int num;
  public:
    TList *next, *prev;
    TList(char *, int);
    ~TList();
    void Draw(int y);
} *beg = NULL, *end = NULL;

TList::TList(char *s, int n)
{
  strcpy(TList::s, s);
  next = NULL;
  prev = NULL;
  num = n;
  if (beg == NULL)
    beg = this;
  else
    end -> next = this;
  prev = end;
  end = this;
}

TList::~TList()
{
}

void TList::Draw(int y)
{
  char t[2];
  t[1] = 0;
  char u[8];
  setfillstyle(1, 3);

  for (int i = 0; i < 3; i++)
  {
    bar(50, y + 30, 80, y + 40);
    ostrstream(u, 6) << num << '.' << ends;
    setcolor(0);
    outtextxy(50, y + 30, u);

    for (int j = 0; j < 3; j++)
    {
      for (int k = 0; k < 3; k++)
      {
        t[0] = s[i * 9 + j * 3 + k];
        int xx = 106 + i * 70 + k * 20;
        int yy = y + 36 + j * 20;
        bar(xx, yy, xx + 8, yy + 8);
        outtextxy(xx, yy, t);
      }
    }
  }

  DrawBigCube(350, y + 40, s);
}

void Error(const char *msg, int exit_code)
{
  cout << "Error: " << msg << '\n';
  exit(exit_code);
}

void Check(char *s)
{
  if (strlen(s) != 27) Error(error1, 4);

  for (int i = 0; i < 27; i++)
    if ((s[i] < 'a') || (s[i] > 'g')) Error(error1, 4);
}

void Read(char *filename)
{
  ifstream in(filename);
  if (!in) Error(error3, 3);

  char *s = new char[255];

  in >> s;
  while (in && s[0])
  {
    Check(s);
    TList *l = new TList(s, ++n);
    l;
    in >> s;
  }

  delete s;
  in.close();
}

void Draw(TList *l)
{
  for (int i = 0; i < 4; i++)
  {
    l -> Draw(i * 120);
    l = l -> next;
  }
  while (kbhit()) getch();
}

void Run()
{
  char c;
  TList *l = beg;
  int i;

  Draw(l);
  do
  {
    c = getch();
    switch (c)
    {
      case 80: if (l -> next -> next -> next -> next != NULL)
               {
                 l = l -> next;
                 Draw(l);
               }
               break;
      case 72: if (l -> prev != NULL)
               {
                 l = l -> prev;
                 Draw(l);
               }
               break;
      case 73: i = 0;
               while ((l -> prev != NULL) && (i < 4))
               {
                 l = l -> prev;
                 i++;
               }
               if (i) Draw(l);
               break;
      case 81: if (l -> next -> next -> next == end) break;
               i = 0;
               while ((l -> next != NULL) && (i < 7))
               {
                 l = l -> next;
                 i++;
               }
               l = l -> prev -> prev -> prev;
               Draw(l);
               break;
      case 71: if (l == beg) break;
               l = beg;
               Draw(l);
               break;
      case 79: if (l -> next -> next -> next == end) break;
               l = end;
               l = l -> prev -> prev -> prev;
               Draw(l);
               break;
    }
  } while (c != 27);
}

void PrintHelp()
{
  cout << "\n\tVIEW.EXE [drive:][path]<file_name>\n";
}

void my_handler()
{
  Error(error4, 2);
}

void DrawAll()
{
  char s[20];

  Draw3DBox(0, 0, 450, getmaxy());
  Draw3DHole(5, 5, 440, getmaxy() - 10, 3);
  setcolor(0);
  for (int u = 0; u < 4; u++)
  {
    int y = u * 120;
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < 4; j++)
      {
        line(100 + i * 70, y + 30 + j * 20, 160 + i * 70, y + 30 + j * 20);
        line(100 + i * 70 + j * 20, y + 30, 100 + i * 70 + j * 20, y + 90);
      }
  }
  setcolor(11);
  rectangle(470, 10, 622, 128);
  outtextxy(480, 20, " - Move Up");
  outtextxy(480, 35, " - Move Down");
  outtextxy(480, 50, "PgUp - Prev page");
  outtextxy(480, 65, "PgDn - Next page");
  outtextxy(480, 80, "Home - First page");
  outtextxy(480, 95, "End - Last page");
  outtextxy(480, 110, "Esc - Exit");
  ostrstream(s, 19) << n << " records read." << ends;
  outtextxy(480, getmaxy() - 30, s);
}

int main(int argc, char **argv)
{
  int gd = DETECT, gm;
  set_new_handler(my_handler);

  if (argc < 2)
  {
    PrintHelp();
    return 0;
  }

  Read(argv[1]);
  registerbgidriver(EGAVGA_driver);
  initgraph(&gd, &gm, "");
  if ((int)graphresult != grOk && (int)graphresult != 781)
    Error(error2, 1);

  DrawAll();
  Run();

  closegraph();
  return 0;
}
